/***************************************************************************
 $RCSfile: rsacard.c,v $
                             -------------------
    cvs         : $Id: rsacard.c,v 1.7 2005/01/30 13:13:05 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "rsacard_p.h"

#include <chipcard2-client/cards/processorcard.h>

#include <aqbanking/banking.h>
#include <aqhbci/mediumrdh.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/misc.h>
#include <gwenhywfar/padd.h>
#include <gwenhywfar/md.h>
#include <gwenhywfar/waitcallback.h>

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#define I18N(msg) msg



GWEN_INHERIT(AH_MEDIUM, AH_MEDIUM_RSACARD)
GWEN_INHERIT(AH_MEDIUMPROVIDER, AH_MEDIUMPROVIDER_RSACARD)


AH_MEDIUM_RSACARD *AH_MediumRSACard_Data_new(){
  AH_MEDIUM_RSACARD *mrdh;

  GWEN_NEW_OBJECT(AH_MEDIUM_RSACARD, mrdh);
  return mrdh;
}



void AH_MediumRSACard_Data_free(AH_MEDIUM_RSACARD *mrdh){
  if (mrdh) {
    GWEN_DB_Group_free(mrdh->dbContexts);
    GWEN_CryptKey_free(mrdh->localSignKey);
    GWEN_CryptKey_free(mrdh->localCryptKey);
    GWEN_CryptKey_free(mrdh->remoteSignKey);
    GWEN_CryptKey_free(mrdh->remoteCryptKey);

    free(mrdh);
  }
}



AH_MEDIUM *AH_MediumRSACard_new(AH_HBCI *hbci,
                                AH_MEDIUMPROVIDER *mp,
                                const char *mediumName) {
  AH_MEDIUM *m;
  AH_MEDIUM_RSACARD *mrdh;

  m=AH_MediumRDH_new(hbci,
                     AH_MEDIUM_RSACARD_NAME,
                     mediumName);
  mrdh=AH_MediumRSACard_Data_new();
  GWEN_INHERIT_SETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD,
		       m, mrdh,
		       AH_MediumRSACard_FreeData);
  mrdh->pluginProvider=mp;

  AH_Medium_SetDeviceType(m, AH_MediumDeviceChipcard);

  /* set implemented functions for Medium */
  AH_Medium_SetMountFn(m, AH_MediumRSACard_Mount);
  AH_Medium_SetCreateFn(m, AH_MediumRSACard_Create);
  AH_Medium_SetUnmountFn(m, AH_MediumRSACard_Unmount);
  AH_Medium_SetGetNextSignSeqFn(m, AH_MediumRSACard_GetNextSignSeq);
  AH_Medium_SetSignFn(m, AH_MediumRSACard_Sign);
  AH_Medium_SetVerifyFn(m, AH_MediumRSACard_Verify);
  AH_Medium_SetEncryptKeyFn(m, AH_MediumRSACard_EncryptKey);
  AH_Medium_SetDecryptKeyFn(m, AH_MediumRSACard_DecryptKey);
  AH_Medium_SetGenerateMsgKeyFn(m, AH_MediumRSACard_GenerateMsgKey);
  AH_Medium_SetSelectContextFn(m, AH_MediumRSACard_SelectContext);
  AH_Medium_SetCreateContextFn(m, AH_MediumRSACard_CreateContext);
  AH_Medium_SetReadContextFn(m, AH_MediumRSACard_ReadContext);
  AH_Medium_SetWriteContextFn(m, AH_MediumRSACard_WriteContext);
  AH_Medium_SetToDbFn(m, AH_MediumRSACard_toDb);
  AH_Medium_SetFromDbFn(m, AH_MediumRSACard_fromDb);


  /* set implemented functions for MediumRDH */
  AH_MediumRDH_SetCreateKeysFn(m, AH_MediumRSACard_CreateKeys);
  AH_MediumRDH_SetActivateKeysFn(m, AH_MediumRSACard_ActivateKeys);

  AH_MediumRDH_SetGetLocalPubSignKeyFn(m, AH_MediumRSACard_GetLocalPubSignKey);
  AH_MediumRDH_SetGetLocalPubCryptKeyFn(m, AH_MediumRSACard_GetLocalPubCryptKey);

  AH_MediumRDH_SetSetPubSignKeyFn(m, AH_MediumRSACard_SetPubSignKey);
  AH_MediumRDH_SetGetPubSignKeyFn(m, AH_MediumRSACard_GetPubSignKey);
  AH_MediumRDH_SetSetPubCryptKeyFn(m, AH_MediumRSACard_SetPubCryptKey);
  AH_MediumRDH_SetGetPubCryptKeyFn(m, AH_MediumRSACard_GetPubCryptKey);

  return m;
}



void AH_MediumRSACard_FreeData(void *bp, void *p){
  AH_MEDIUM *m;
  AH_MEDIUM_RSACARD *mrdh;

  assert(bp);
  assert(p);
  m=(AH_MEDIUM*)bp;
  mrdh=(AH_MEDIUM_RSACARD*)p;
  AH_MediumRSACard_Data_free(mrdh);
}



int AH_MediumRSACard_Mount(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  LC_CLIENT_RESULT res;
  GWEN_DB_NODE *dbCardData;
  LC_CLIENT *client;
  LC_CARD *hcard;
  int first;
  int havepin;
  const char *currCardNumber;
  int herr;
  GWEN_TYPE_UINT32 bid;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  mrdh->hasEgPin=0;

  client=AH_MediumRSACardProvider_GetClient(mrdh->pluginProvider);
  assert(client);

  bid=AB_Banking_ShowBox(AH_Medium_GetBankingApi(m),
			 0,
			 "Accessing Card",
                         "Mounting card, please wait...");

  res=LC_Client_StartWait(client, 0, 0);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not send StartWait request");
    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
    return -1;
  }

  first=1;
  for (;;) {
    int timeout;

    // determine timeout value
    if (first)
      timeout=3;
    else
      timeout=5;

    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			   0,
			   AB_Banking_LogLevelInfo,
			   I18N("Waiting for card"));
    hcard=LC_Client_WaitForNextCard(client, timeout);
    if (!hcard) {
      int mres;

      mres=AH_Medium_UI(m, AH_Medium_UIActionInsertMedium);
      if (mres) {
	DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
	LC_Client_StopWait(client);
	AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
	return -1;
      }
    }
    else {
      /* ok, we have a card, now check it */
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			     0,
			     AB_Banking_LogLevelInfo,
			     I18N("Card found"));
      if (LC_Starcos_ExtendCard(hcard)) {
        DBG_ERROR(AQHBCI_LOGDOMAIN,
                  "RSA card not available, please check your setup");
        LC_Card_free(hcard);
        LC_Client_StopWait(client);
	AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
        return -1;
      }

      res=LC_Card_Open(hcard);
      if (res!=LC_Client_ResultOk) {
        LC_Card_free(hcard);
        DBG_NOTICE(AQHBCI_LOGDOMAIN,
                   "Could not open card, maybe not a RSA card?");
      } /* if card not open */
      else {
        const char *mname;

        mname=AH_Medium_GetMediumName(m);
        dbCardData=LC_Starcos_GetCardDataAsDb(hcard);
	assert(dbCardData);

        currCardNumber=GWEN_DB_GetCharValue(dbCardData,
                                            "iccsn/cardNumber",
                                            0,
                                            0);
	if (!currCardNumber) {
          DBG_ERROR(AQHBCI_LOGDOMAIN,
                    "INTERNAL: No card number in card data.");
	  abort();
	}

        DBG_NOTICE(AQHBCI_LOGDOMAIN, "Card number: %s", currCardNumber);

        if (!mname) {
	  DBG_NOTICE(AQHBCI_LOGDOMAIN, "No medium name");
          AH_Medium_SetMediumName(m, currCardNumber);
          mname=AH_Medium_GetMediumName(m);
          break;
        }
        if (!*mname) {
          DBG_NOTICE(AQHBCI_LOGDOMAIN, "Empty medium name");
	  AH_Medium_SetMediumName(m, currCardNumber);
          mname=AH_Medium_GetMediumName(m);
          break;
        }

        if (strcasecmp(mname, currCardNumber)==0) {
          DBG_NOTICE(AQHBCI_LOGDOMAIN, "Card number equals");
          break;
        }

        LC_Card_Close(hcard);
	LC_Card_free(hcard);

        hcard=LC_Client_PeekNextCard(client);
        if (!hcard) {
	  int mres;
    
	  mres=AH_Medium_UI(m, AH_Medium_UIActionInsertCorrectMedium);
	  if (mres) {
	    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
	    LC_Client_StopWait(client);
	    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
	    return -1;
	  }
	} /* if there is no other card waiting */
        else {
          /* otherwise there already is another card in another reader,
           * so no need to bother the user. This allows to insert all
           * cards in all readers and let me choose the card ;-) */
        } /* if there is another card waiting */
      } /* if card open */
    } /* if there is a card */

    first=0;
  } /* for */

  /* ok, now we have the card we wanted to have, now ask for the pin */
  LC_Client_StopWait(client);
  havepin=0;
  while(!havepin) {
    /* enter pin */
    herr=AH_MediumRSACard__enterPin(m, hcard, 0x90, 0);
    if (herr) {
      LC_Card_Close(hcard);
      LC_Card_free(hcard);
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in PIN input");
      AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
      return -1;
    }
    else
      havepin=1;
  } /* while !havepin */

  /* ok, now we have a card and the PIN is verified, that's it ;-) */
  mrdh->card=hcard;

  AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);

  /* read contexts */
  if (mrdh->dbContexts==0) {
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			   0,
			   AB_Banking_LogLevelNotice,
			   I18N("First mount, will cache some data, "
				"please wait"));
    herr=AH_MediumRSACard__readContexts(m);
    if (herr) {
      mrdh->card=0;
      LC_Card_Close(hcard);
      LC_Card_free(hcard);
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error reading key data");
      AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
      return -1;
    }
  }

  if (!AH_Medium_GetMediumName(m) && currCardNumber) {
    AH_Medium_SetMediumName(m, currCardNumber);
  }

  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Medium mounted");
  return 0;
}



void AH_MediumRSACard__showError(LC_CARD *card, LC_CLIENT_RESULT res,
                                 const char *x) {
  const char *s;

  switch(res) {
  case LC_Client_ResultOk:
    s="Ok.";
    break;
  case LC_Client_ResultWait:
    s="Timeout.";
    break;
  case LC_Client_ResultIpcError:
    s="IPC error.";
    break;
  case LC_Client_ResultCmdError:
    s="Command error.";
    break;
  case LC_Client_ResultDataError:
    s="Data error.";
    break;
  case LC_Client_ResultAborted:
    s="Aborted.";
    break;
  case LC_Client_ResultInvalid:
    s="Invalid argument to command.";
    break;
  case LC_Client_ResultInternal:
    s="Internal error.";
    break;
  case LC_Client_ResultGeneric:
    s="Generic error.";
    break;
  default:
    s="Unknown error.";
    break;
  }

  DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in \"%s\": %s", x, s);
  if (res==LC_Client_ResultCmdError) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "  Last card command result:\n");
    DBG_ERROR(AQHBCI_LOGDOMAIN, "   SW1=%02x, SW2=%02x\n",
              LC_Card_GetLastSW1(card),
              LC_Card_GetLastSW2(card));
    s=LC_Card_GetLastResult(card);
    if (s) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "   Result: %s\n", s);
    }
    s=LC_Card_GetLastText(card);
    if (s) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "   Text  : %s\n", s);
    }
  }
}



int AH_MediumRSACard_InputDevicePin(AH_MEDIUM *m,
                                    char *pwbuffer,
                                    int minLen,
                                    int maxLen,
                                    int flags){
  int rv;
  const char *name;
  char buffer[512];

  assert(m);

  buffer[0]=0;
  buffer[sizeof(buffer)-1]=0;
  if (flags & AB_BANKING_INPUT_FLAGS_CONFIRM) {
    snprintf(buffer, sizeof(buffer)-1,
             I18N("Please enter a new device PIN for \n"
                  "%s\n"
                  "The PIN must be at least %d digits long.\n"
                  "You must only enter numbers, not letters.\n"
                  "\n"
                  "PLEASE NOTE that the device PIN is a special PIN,\n"
                  "it is especially NOT your normal cardholder PIN !\n"
                  "If unsure please abort this operation.\n"
                  "<html>"
                  "<p>Please enter a new device PIN for <i>%s</i>.</p>"
                  "<p>This PIN must be at least %d digits long.</p>"
                  "<p><font colour=\"red\">"
                  "Please note that the device PIN is a special PIN,"
                  "it is especially <b>not</b> your normal cardholder "
                  "PIN !</p>"
                  "<p>If unsure please abort this operation.</p>"
                  "</font></p>"
                  "</html>"),
             AH_Medium_GetDescriptiveName(m),
             minLen,
             AH_Medium_GetDescriptiveName(m),
             minLen);
  }
  else {
    snprintf(buffer, sizeof(buffer)-1,
             I18N("Please enter the device PIN for \n"
                  "%s\n"
                  "The PIN must be at least %d digits long.\n"
                  "You must only enter numbers, not letters.\n"
                  "\n"
                  "PLEASE NOTE that the device PIN is a special PIN,\n"
                  "it is especially NOT your normal cardholder PIN !\n"
                  "If unsure please abort this operation.\n"
                  "<html>"
                  "<p>Please enter the device PIN for <i>%s</i>.</p>"
                  "<p>This PIN must be at least %d digits long.</p>"
                  "<p><font colour=\"red\">"
                  "Please note that the device PIN is a special PIN,"
                  "it is especially <b>not</b> your normal cardholder "
                  "PIN !</p>"
                  "<p>If unsure please abort this operation.</p>"
                  "</font></p>"
                  "</html>"),
             AH_Medium_GetDescriptiveName(m),
             minLen,
             AH_Medium_GetDescriptiveName(m),
             minLen);
  }

  name=AH_Medium_GetMediumName(m);
  if (name) {
    GWEN_BUFFER *nbuf;

    nbuf=GWEN_Buffer_new(0, 256 ,0 ,1);
    GWEN_Buffer_AppendString(nbuf, "EGPIN::");
    GWEN_Buffer_AppendString(nbuf, name);
    rv=AB_Banking_GetPin(AH_Medium_GetBankingApi(m),
                         flags,
                         GWEN_Buffer_GetStart(nbuf),
                         I18N("Enter Device PIN"),
			 buffer,
			 pwbuffer,
                         minLen,
                         maxLen);
    GWEN_Buffer_free(nbuf);
  }
  else {
    rv=AB_Banking_InputBox(AH_Medium_GetBankingApi(m),
                           flags,
                           I18N("Enter Device PIN"),
                           buffer,
			   pwbuffer,
                           minLen,
                           maxLen);
  }
  return rv;
}



int AH_MediumRSACard__enterPin(AH_MEDIUM *m, LC_CARD *hcard,
                               int pid, int iniPin){
  LC_CLIENT_RESULT res;
  int maxErrors;
  int currentErrors;

  res=LC_Starcos_GetPinStatus(hcard, pid, &maxErrors, &currentErrors);
  if (res==LC_Client_ResultCmdError) {
    if (LC_Card_GetLastSW1(hcard)==0x69 &&
	LC_Card_GetLastSW2(hcard)==0x85) {
      /* still initial PIN */
      DBG_WARN(AQHBCI_LOGDOMAIN, "Still initial PIN");
      /* TODO: ask user for initial PIN change */
      AH_MediumRSACard__showError(hcard, res, "GetPinStatus");
      return -1;
    }
    else {
      AH_MediumRSACard__showError(hcard, res, "GetPinStatus");
      return -1;
    }
  }
  else if (res!=LC_Client_ResultOk) {
    AH_MediumRSACard__showError(hcard, res, "GetPinStatus");
    return -1;
  }

  if (currentErrors!=maxErrors) {
    int rv;

    /* Ask for force */
    rv=AB_Banking_MessageBox
      (AH_Medium_GetBankingApi(m),
       AB_BANKING_MSG_FLAGS_TYPE_WARN |
       AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
       AB_BANKING_MSG_FLAGS_CONFIRM_B1,
       I18N("Warning"),
       I18N("At least once a false PIN has been entered.\n"
	    "You should be very carefull about PIN entry, if\n"
	    "you enter a bad PIN too often the card gets blocked.\n"
	    "\n"
	    "Are you sufficiently sure about the PIN to enter it now ?"
	    "<html>"
	    "<p>At least once a false PIN has been entered.</p>"
	    "<p>You should be very carefull about PIN entry, if "
	    "you enter a bad PIN too often the card gets <b>blocked</b>.</p>"
	    "<p>Are you sufficiently sure about the PIN to enter it now ?</p>"
	    "</html>"
	   ),
       I18N("Yes"), I18N("Abort"), 0);
    if (rv!=1) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "User aborted.");
      return AB_ERROR_USER_ABORT;
    }
  }

  /* device pin */
  if (iniPin) {
    res=LC_Starcos_VerifyInitialPin(hcard, pid);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(hcard, res, "VerifyInitialPin");
    }
  }
  else {
    if (LC_Card_GetReaderFlags(hcard) & LC_CARD_READERFLAGS_KEYPAD) {
      int mres;

      DBG_INFO(AQHBCI_LOGDOMAIN, "Terminal has a keypad, will ask for pin.");
      /* tell the user about pin verification */
      mres=AH_Medium_UI(m, AH_Medium_UIActionStartInputPinViaKeypad);
      if (mres) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
        return -1;
      }
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                             0,
                             AB_Banking_LogLevelNotice,
                             "Secure PIN input");
      res=LC_Starcos_SecureVerifyPin(hcard, pid);
  
      /* tell the user about end of pin verification */
      AH_Medium_UI(m, AH_Medium_UIActionFinishedInputPinViaKeypad);

      if (res!=LC_Client_ResultOk) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                  LC_Card_GetLastSW1(hcard),
                  LC_Card_GetLastSW2(hcard),
                  LC_Card_GetLastText(hcard));
  
        if (LC_Card_GetLastSW1(hcard)==0x63) {
          switch (LC_Card_GetLastSW2(hcard)) {
          case 0xc0:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelEmergency,
                                   "Bad PIN, no error left");
            AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                  AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                  AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                  AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                  "Serious Warning",
                                  "The PIN you entered is invalid, there is "
                                  "no room left for errors.<br>"
                                  "It is <b>strongly</b> recommended to "
                                  "temporarily<br>"
                                  "stop using this current program.<br>"
                                  "Please use the last program which "
                                  "successfully verified the PIN.<br>"
                                  "This will reset the card's error "
                                  "counter.<br>"
                                  "After that you may come back to this "
                                  "program",
                                  "Ok", 0, 0);
            break;
          case 0xc1:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN, one error left");
            break;
          case 0xc2:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN, two errors left");
            break;
          default:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN");
            break;
          } // switch
          return -1;
        }
        else if (LC_Card_GetLastSW1(hcard)==0x69 &&
                 LC_Card_GetLastSW2(hcard)==0x83) {
          DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Card unusable");
          return -1;
        }
        else if (LC_Card_GetLastSW1(hcard)==0x64 &&
                 LC_Card_GetLastSW2(hcard)==0x01) {
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Aborted by user");
          return AB_ERROR_USER_ABORT;
        }
        else {
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Unknown error");
          return -1;
        }
      } /* if not ok */
      else {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelNotice,
                               "PIN ok.");
      }
    }
    else {
      char pinBuffer[64];
      int iprv;

      DBG_INFO(AQHBCI_LOGDOMAIN, "No keypad (or disabled), will ask for PIN");
      memset(pinBuffer, 0, sizeof(pinBuffer));

      if (pid==0x90)
        iprv=AH_Medium_InputPin(m,
                                pinBuffer,
                                AH_MEDIUM_RSACARD_PINMINLENGTH,
                                sizeof(pinBuffer),
                                AB_BANKING_INPUT_FLAGS_NUMERIC);
      else
        iprv=AH_MediumRSACard_InputDevicePin(m,
                                             pinBuffer,
                                             AH_MEDIUM_RSACARD_PINMINLENGTH,
                                             sizeof(pinBuffer),
                                             AB_BANKING_INPUT_FLAGS_NUMERIC);
      if (iprv) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Error asking for PIN, aborting");
        memset(pinBuffer, 0, sizeof(pinBuffer));
        return AB_ERROR_ABORTED;
      }

      DBG_INFO(AQHBCI_LOGDOMAIN, "Verifying the PIN");
      res=LC_Starcos_VerifyPin(hcard, pid, pinBuffer);
      if (res!=LC_Client_ResultOk) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                  LC_Card_GetLastSW1(hcard),
                  LC_Card_GetLastSW2(hcard),
                  LC_Card_GetLastText(hcard));
    
        if (LC_Card_GetLastSW1(hcard)==0x63) {
          AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
          switch (LC_Card_GetLastSW2(hcard)) {
          case 0xc0:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelEmergency,
                                   "Bad PIN, no error left");
            AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                  AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                  AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                  AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                  "Serious Warning",
                                  "The PIN you entered is invalid, there is "
                                  "no room left for errors.<br>"
                                  "It is <b>strongly</b> recommended to "
                                  "temporarily<br>"
                                  "stop using this current program.<br>"
                                  "Please use the last program which "
                                  "successfully verified the PIN.<br>"
                                  "This will reset the card's error "
                                  "counter.<br>"
                                  "After that you may come back to this "
                                  "program",
                                  "Ok", 0, 0);
            break;
          case 0xc1:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN, one error left");
            break;
          case 0xc2:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN, two errors left");
            break;
          default:
            AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                   0,
                                   AB_Banking_LogLevelError,
                                   "Bad PIN");
            break;
          } // switch
          return -1;
        }
        else if (LC_Card_GetLastSW1(hcard)==0x69 &&
                 LC_Card_GetLastSW2(hcard)==0x83) {
          AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
          DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Card unusable");
          return -1;
        }
        else if (LC_Card_GetLastSW1(hcard)==0x64 &&
                 LC_Card_GetLastSW2(hcard)==0x01) {
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Aborted by user");
          return AB_ERROR_USER_ABORT;
        }
        else {
          DBG_ERROR(AQHBCI_LOGDOMAIN, "Unknown error");
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Unknown error");
          return -1;
        }
      } // if not ok
      else {
        DBG_INFO(AQHBCI_LOGDOMAIN, "PIN ok");
        AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusOk);
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelNotice,
                               "PIN ok.");
      }
    }
  }


  return 0;






  if (LC_Card_GetReaderFlags(hcard) & LC_CARD_READERFLAGS_KEYPAD) {
    int mres;

    DBG_INFO(AQHBCI_LOGDOMAIN,"Terminal has a keypad, will ask for pin.");
    /* tell the user about pin verification */
    mres=AH_Medium_UI(m, AH_Medium_UIActionStartInputPinViaKeypad);
    if (mres) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
      return -1;
    }
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                           0,
                           AB_Banking_LogLevelNotice,
                           "Secure PIN input");
    res=LC_Starcos_SecureVerifyPin(hcard, pid);

    /* tell the user about end of pin verification */
    mres=AH_Medium_UI(m, AH_Medium_UIActionFinishedInputPinViaKeypad);
    if (mres) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in user interaction");
      return -1;
    }

    if (res!=LC_Client_ResultOk) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                LC_Card_GetLastSW1(hcard),
                LC_Card_GetLastSW2(hcard),
                LC_Card_GetLastText(hcard));

      if (LC_Card_GetLastSW1(hcard)==0x63) {
	switch (LC_Card_GetLastSW2(hcard)) {
        case 0xc0:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelEmergency,
                                 "Bad PIN, no error left");
          AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                "Serious Warning",
                                "The PIN you entered is invalid, there is "
                                "no room left for errors.<br>"
                                "It is <b>strongly</b> recommended to "
                                "temporarily<br>"
                                "stop using this current program.<br>"
                                "Please use the last program which "
                                "successfully verified the PIN.<br>"
                                "This will reset the card's error "
                                "counter.<br>"
                                "After that you may come back to this "
                                "program",
                                "Ok", 0, 0);
          break;
        case 0xc1:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN, one error left");
          break;
        case 0xc2:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN, two errors left");
          break;
        default:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN");
          break;
        } // switch
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
               LC_Card_GetLastSW2(hcard)==0x83) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Card unusable");
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x64 &&
               LC_Card_GetLastSW2(hcard)==0x01) {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Aborted by user");
        return -1;
      }
      else {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Unknown error");
        return -1;
      }
    } /* if not ok */
    else {
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                             0,
                             AB_Banking_LogLevelNotice,
                             "PIN ok.");
    }
  } /* if hasKeyPad */
  else {
    char pinBuffer[64];

    DBG_INFO(AQHBCI_LOGDOMAIN, "No keypad (or disabled), will ask for PIN");
    memset(pinBuffer, 0, sizeof(pinBuffer));
    if (AH_Medium_InputPin(m,
			   pinBuffer,
			   AH_MEDIUM_RSACARD_PINMINLENGTH,
			   sizeof(pinBuffer),
			   AB_BANKING_INPUT_FLAGS_NUMERIC)) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "Error asking for PIN, aborting");
      memset(pinBuffer, 0, sizeof(pinBuffer));
      return -1;
    }

    DBG_INFO(AQHBCI_LOGDOMAIN, "Verifying the PIN");

    res=LC_Starcos_VerifyPin(hcard, pid, pinBuffer);
    if (res!=LC_Client_ResultOk) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "sw1=%02x sw2=%02x (%s)",
                LC_Card_GetLastSW1(hcard),
                LC_Card_GetLastSW2(hcard),
                LC_Card_GetLastText(hcard));

      if (LC_Card_GetLastSW1(hcard)==0x63) {
	AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
	switch (LC_Card_GetLastSW2(hcard)) {
        case 0xc0:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelEmergency,
                                 "Bad PIN, no error left");
          AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                                AB_BANKING_MSG_FLAGS_TYPE_WARN |
                                AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                                AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                                "Serious Warning",
                                "The PIN you entered is invalid, there is "
                                "no room left for errors.<br>"
                                "It is <b>strongly</b> recommended to "
                                "temporarily<br>"
                                "stop using this current program.<br>"
                                "Please use the last program which "
                                "successfully verified the PIN.<br>"
                                "This will reset the card's error "
                                "counter.<br>"
                                "After that you may come back to this "
                                "program",
                                "Ok", 0, 0);
          break;
        case 0xc1:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN, one error left");
          break;
        case 0xc2:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN, two errors left");
          break;
        default:
          AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                                 0,
                                 AB_Banking_LogLevelError,
                                 "Bad PIN");
          break;
        } // switch
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
               LC_Card_GetLastSW2(hcard)==0x83) {
	AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusBad);
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Card unusable");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Card unusable");
        return -1;
      }
      else if (LC_Card_GetLastSW1(hcard)==0x64 &&
               LC_Card_GetLastSW2(hcard)==0x01) {
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Aborted by user");
        return -1;
      }
      else {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Unknown error");
        AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                               0,
                               AB_Banking_LogLevelError,
                               "Unknown error");
        return -1;
      }
    } // if not ok
    else {
      DBG_INFO(AQHBCI_LOGDOMAIN, "PIN ok");
      AH_Medium_SetPinStatus(m, pinBuffer, AB_Banking_PinStatusOk);
      AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
                             0,
                             AB_Banking_LogLevelNotice,
                             "PIN ok.");
    }
  } // if no keyPad

  return 0;
}



int AH_MediumRSACard__readContexts(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbContexts;
  GWEN_DB_NODE *dbCurr;
  LC_CLIENT_RESULT res;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  dbContexts=mrdh->dbContexts;
  if (!dbContexts) {
    GWEN_TYPE_UINT32 bid;
    int idx;

    bid=AB_Banking_ShowBox(AH_Medium_GetBankingApi(m),
			   0,
			   I18N("Accessing Card"),
			   I18N("Reading card, please wait..."));
    dbContexts=GWEN_DB_Group_new("contexts");
    DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading institute data");
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			   0,
			   AB_Banking_LogLevelNotice,
			   I18N("Reading institute data"));
    res=LC_Starcos_ReadInstituteData(mrdh->card, 0, dbContexts);
    if (res!=LC_Client_ResultOk) {
      if (res==LC_Client_ResultNoData) {
	DBG_WARN(AQHBCI_LOGDOMAIN, "No context available");
      }
      else {
	GWEN_DB_Group_free(dbContexts);
	AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
	return -1;
      }
    }

    dbCurr=GWEN_DB_FindFirstGroup(dbContexts, "context");
    idx=0;
    AB_Banking_ProgressLog(AH_Medium_GetBankingApi(m),
			   0,
			   AB_Banking_LogLevelNotice,
			   I18N("Reading public keys"));
    while(dbCurr) {
      GWEN_DB_NODE *gr;
      GWEN_CRYPTKEY *key;
      GWEN_KEYSPEC *ks;
      int kid;
      const char *userId;
      const char *bankId;

      userId=GWEN_DB_GetCharValue(dbCurr, "userId", 0, "");
      bankId=GWEN_DB_GetCharValue(dbCurr, "bankId", 0, userId);

      /* read user keys */
      gr=GWEN_DB_GetGroup(dbCurr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                          "userKeys");
      assert(gr);

      /* read local sign key */
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading local sign key");
      kid=0x81+idx;
      ks=LC_Starcos_GetKeySpec(mrdh->card, kid);
      assert(ks);
      if (GWEN_KeySpec_GetStatus(ks)==0x08) {
        DBG_NOTICE(AQHBCI_LOGDOMAIN, "No local sign key");
      }
      else {
        key=LC_Starcos_ReadPublicKey(mrdh->card, kid);
        if (key) {
          GWEN_DB_NODE *dbKey;
          GWEN_ERRORCODE err;
  
	  GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
	  GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
	  GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
          GWEN_CryptKey_SetOwner(key, userId);
          dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                                 "signKey");
          err=GWEN_CryptKey_ToDb(key, dbKey, 1);
          if (!GWEN_Error_IsOk(err)) {
            DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
            GWEN_KeySpec_free(ks);
            AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
            return -1;
          }
        }
      }
      GWEN_KeySpec_free(ks);

      /* read local crypt key */
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading local crypt key");
      kid=0x86+idx;
      ks=LC_Starcos_GetKeySpec(mrdh->card, kid);
      assert(ks);
      if (GWEN_KeySpec_GetStatus(ks)==0x08) {
        DBG_NOTICE(AQHBCI_LOGDOMAIN, "No local crypt key");
      }
      else {
        key=LC_Starcos_ReadPublicKey(mrdh->card, kid);
        if (key) {
          GWEN_DB_NODE *dbKey;
          GWEN_ERRORCODE err;

	  GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
	  GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
	  GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
          GWEN_CryptKey_SetOwner(key, userId);
          dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                                 "cryptKey");
          err=GWEN_CryptKey_ToDb(key, dbKey, 1);
          if (!GWEN_Error_IsOk(err)) {
            DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
            GWEN_KeySpec_free(ks);
            AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
            return -1;
          }
        }
      }
      GWEN_KeySpec_free(ks);

      /* read bank keys */
      gr=GWEN_DB_GetGroup(dbCurr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
			  "bankKeys");
      assert(gr);

      /* read remote sign key */
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading remote sign key");
      kid=0x91+idx;
      ks=LC_Starcos_GetKeySpec(mrdh->card, kid);
      assert(ks);
      if (GWEN_KeySpec_GetStatus(ks)==0x08) {
        DBG_NOTICE(AQHBCI_LOGDOMAIN, "No remote sign key");
      }
      else {
        key=LC_Starcos_ReadPublicKey(mrdh->card, kid);
        if (key) {
          GWEN_DB_NODE *dbKey;
          GWEN_ERRORCODE err;

	  GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
	  GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
	  GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
          GWEN_CryptKey_SetOwner(key, userId);
          dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                                 "signKey");
          err=GWEN_CryptKey_ToDb(key, dbKey, 1);
          if (!GWEN_Error_IsOk(err)) {
            DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
            GWEN_KeySpec_free(ks);
            AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
            return -1;
          }
        }
      }
      GWEN_KeySpec_free(ks);

      /* read remote crypt key */
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading remote crypt key");
      kid=0x96+idx;
      ks=LC_Starcos_GetKeySpec(mrdh->card, kid);
      assert(ks);
      if (GWEN_KeySpec_GetStatus(ks)==0x08) {
        DBG_NOTICE(AQHBCI_LOGDOMAIN, "No remote sign key");
      }
      else {
        key=LC_Starcos_ReadPublicKey(mrdh->card, kid);
        if (key) {
          GWEN_DB_NODE *dbKey;
          GWEN_ERRORCODE err;

	  GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
	  GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
	  GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
          GWEN_CryptKey_SetOwner(key, userId);
          dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                                 "cryptKey");
          err=GWEN_CryptKey_ToDb(key, dbKey, 1);
          if (!GWEN_Error_IsOk(err)) {
            DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
            GWEN_KeySpec_free(ks);
            AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
            return -1;
          }
        }
      }
      GWEN_KeySpec_free(ks);

      dbCurr=GWEN_DB_FindNextGroup(dbCurr, "context");
      idx++;
    }
    mrdh->dbContexts=dbContexts;

    AB_Banking_HideBox(AH_Medium_GetBankingApi(m), bid);
  }

  return 0;
}


int AH_MediumRSACard_Create(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  rv=AH_MediumRSACard_Mount(m);
  if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  return 0;
}



int AH_MediumRSACard_Unmount(AH_MEDIUM *m, int force){
  AH_MEDIUM_RSACARD *mrdh;
  LC_CLIENT_RESULT res;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  mrdh->selected=0;
  mrdh->account=0;

  GWEN_CryptKey_free(mrdh->localSignKey);
  mrdh->localSignKey=0;
  GWEN_CryptKey_free(mrdh->localCryptKey);
  mrdh->localCryptKey=0;
  GWEN_CryptKey_free(mrdh->remoteSignKey);
  mrdh->remoteSignKey=0;
  GWEN_CryptKey_free(mrdh->remoteCryptKey);
  mrdh->remoteCryptKey=0;

  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Closing card");
  res=LC_Card_Close(mrdh->card);
  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Deleting card");
  LC_Card_free(mrdh->card);
  mrdh->card=0;
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not unmount chip card");
    return -1;
  }
  DBG_NOTICE(AQHBCI_LOGDOMAIN, "Medium unmounted");
  return 0;
}



AH_MEDIUM_RESULT AH_MediumRSACard_Sign(AH_MEDIUM *m,
				       GWEN_BUFFER *msgbuf,
				       GWEN_BUFFER *signbuf){
  AH_MEDIUM_RSACARD *mrdh;
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *hashbuf;
  char buffer[20];
  unsigned int bsize;
  int kid;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hash data");
  bsize=sizeof(buffer);
  if (GWEN_MD_Hash("RMD160",
                   GWEN_Buffer_GetStart(msgbuf),
                   GWEN_Buffer_GetUsedBytes(msgbuf),
                   buffer,
                   &bsize)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error hashing message");
    return AH_MediumResultGenericError;
  }
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hashing done");

  hashbuf=GWEN_Buffer_new(0, bsize, 0, 1);
  GWEN_Buffer_AppendBytes(hashbuf, buffer, bsize);

  /* sign hash */
  GWEN_Buffer_Rewind(hashbuf);
  kid=0x80+mrdh->account;
  res=LC_Starcos_Sign(mrdh->card, kid, hashbuf, signbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error signing hash (%d)", res);
    GWEN_Buffer_free(hashbuf);
    return AH_MediumResultGenericError;
  }

  GWEN_Buffer_free(hashbuf);
  return AH_MediumResultOk;
}



int AH_MediumRSACard_GetNextSignSeq(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_TYPE_UINT32 seq;
  int kid;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No user selected.");
    return 0;
  }
  kid=0x80+mrdh->account;
  seq=LC_Starcos_ReadSigCounter(mrdh->card, kid);
  if (!seq) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Could not read signature counter.");
    return 0;
  }
  return seq;
}



AH_MEDIUM_RESULT AH_MediumRSACard_Verify(AH_MEDIUM *m,
					 GWEN_BUFFER *msgbuf,
					 GWEN_BUFFER *signbuf,
					 int signseq){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_BUFFER *hashbuf;
  GWEN_ERRORCODE err;
  char buffer[20];
  unsigned int bsize;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (mrdh->remoteSignKey==0) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No remote sign key");
    return AH_MediumResultNoKey;
  }

  /* hash data */
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hashing data");
  bsize=sizeof(buffer);
  if (GWEN_MD_Hash("RMD160",
                   GWEN_Buffer_GetStart(msgbuf),
                   GWEN_Buffer_GetUsedBytes(msgbuf),
                   buffer,
                   &bsize)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error hashing message");
    return AH_MediumResultGenericError;
  }
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Hashing done");

  hashbuf=GWEN_Buffer_new(0, bsize, 0, 1);
  GWEN_Buffer_AppendBytes(hashbuf, buffer, bsize);

  /* padd */
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Padding hash using ISO 9796");
  if (GWEN_Padd_PaddWithISO9796(hashbuf)) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    GWEN_Buffer_free(hashbuf);
    return AH_MediumResultGenericError;
  }

  /* verify hash */
  GWEN_Buffer_Rewind(hashbuf);
  GWEN_Buffer_Rewind(signbuf);
  err=GWEN_CryptKey_Verify(mrdh->remoteSignKey,
                           hashbuf,
                           signbuf);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
    DBG_WARN(AQHBCI_LOGDOMAIN, "Invalid signature");
    GWEN_Buffer_free(hashbuf);
    return AH_MediumResultInvalidSignature;
  }
  DBG_INFO(AQHBCI_LOGDOMAIN, "Signature is valid");

  GWEN_Buffer_free(hashbuf);
  DBG_DEBUG(AQHBCI_LOGDOMAIN, "Verification done");
  return AH_MediumResultOk;


}



GWEN_BUFFER *AH_MediumRSACard_GenerateMsgKey(AH_MEDIUM *m) {
  GWEN_BUFFER *kbuf;
  LC_CLIENT_RESULT res;
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  kbuf=GWEN_Buffer_new(0, 16, 0, 1);

  /* create left */
  res=LC_Starcos_GetChallenge(mrdh->card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting random-1 (%d)", res);
    GWEN_Buffer_free(kbuf);
    return 0;
  }

  /* create right */
  res=LC_Starcos_GetChallenge(mrdh->card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error getting random-2 (%d)", res);
    GWEN_Buffer_free(kbuf);
    return 0;
  }

  return kbuf;
}



AH_MEDIUM_RESULT AH_MediumRSACard_EncryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *encKey){
  AH_MEDIUM_RSACARD *mrdh;
  AH_MEDIUM_RESULT rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  /* check size */
  if (GWEN_Buffer_GetUsedBytes(srckey)!=16) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Bad key length (expected 16, got %d)",
              GWEN_Buffer_GetUsedBytes(srckey));
    return AH_MediumResultBadKey;
  }

  rv=AH_MediumRDH_EncryptKey(m, mrdh->remoteCryptKey, srckey, encKey);
  if (rv) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error encrypting key (%d)", rv);
    return rv;
  }

  return AH_MediumResultOk;
}



AH_MEDIUM_RESULT AH_MediumRSACard_DecryptKey(AH_MEDIUM *m,
                                             GWEN_BUFFER *srckey,
                                             GWEN_BUFFER *dstkey){
  AH_MEDIUM_RSACARD *mrdh;
  int kid;
  LC_CLIENT_RESULT res;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  kid=0x85+mrdh->account;
  res=LC_Starcos_Decipher(mrdh->card, kid, srckey, dstkey);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error decrypting block (%d)", res);
    return AH_MediumResultGenericError;
  }
  return AH_MediumResultOk;
}



int AH_MediumRSACard_SelectContext(AH_MEDIUM *m, int idx) {
  AH_MEDIUM_RSACARD *mrdh;
  const char *secId;
  GWEN_DB_NODE *dbCurr;
  GWEN_DB_NODE *dbKey;
  int i;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    mrdh->selected=0;
    return -1;
  }

  if (idx<0 || idx>AH_MEDIUM_RSACARD_MAXCONTEXT) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "Context index %d out of boundaries",
	      idx);
    return -1;
  }

  if (!mrdh->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  if (mrdh->selected) {
    assert(mrdh->dbCurrentContext);
    assert(mrdh->account);
    if (mrdh->account==idx) {
      /* already selected */
      return 0;
    }
  }

  /* search for context */
  dbCurr=GWEN_DB_FindFirstGroup(mrdh->dbContexts, "context");
  i=idx;
  while(dbCurr && i--)
    dbCurr=GWEN_DB_FindNextGroup(dbCurr, "context");

  if (!dbCurr) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context %d not found", idx);
    return -1;
  }

  /* store context data */
  mrdh->selected=0;
  mrdh->dbCurrentContext=dbCurr;
  mrdh->account=idx;

  GWEN_CryptKey_free(mrdh->localSignKey); mrdh->localSignKey=0;
  GWEN_CryptKey_free(mrdh->localCryptKey); mrdh->localCryptKey=0;
  GWEN_CryptKey_free(mrdh->remoteSignKey); mrdh->remoteSignKey=0;
  GWEN_CryptKey_free(mrdh->remoteCryptKey); mrdh->remoteCryptKey=0;


  /* set security Id */
  secId=GWEN_DB_GetCharValue(dbCurr, "systemId", 0, 0);
  if (secId)
    AH_Medium_SetSecurityId(m, secId, strlen(secId));

  /* set keys, adjust keyspecs */
  /* local sign key */
  dbKey=GWEN_DB_GetGroup(dbCurr, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			 "userKeys/signKey");
  if (dbKey) {
    GWEN_CRYPTKEY *key;

    key=GWEN_CryptKey_FromDb(dbKey);
    if (key) {
      AH_Medium_SetLocalSignKeySpec(m, GWEN_CryptKey_GetKeySpec(key));
      GWEN_CryptKey_free(mrdh->localSignKey);
      mrdh->localSignKey=key;
    }
  }

  /* local crypt key */
  dbKey=GWEN_DB_GetGroup(dbCurr, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			 "userKeys/cryptKey");
  if (dbKey) {
    GWEN_CRYPTKEY *key;

    key=GWEN_CryptKey_FromDb(dbKey);
    if (key) {
      AH_Medium_SetLocalCryptKeySpec(m, GWEN_CryptKey_GetKeySpec(key));
      mrdh->localCryptKey=key;
    }
  }

  /* remote sign key */
  dbKey=GWEN_DB_GetGroup(dbCurr, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			 "bankKeys/signKey");
  if (dbKey) {
    GWEN_CRYPTKEY *key;

    key=GWEN_CryptKey_FromDb(dbKey);
    if (key) {
      AH_Medium_SetRemoteSignKeySpec(m, GWEN_CryptKey_GetKeySpec(key));
      GWEN_CryptKey_free(mrdh->remoteSignKey);
      mrdh->remoteSignKey=key;
    }
  }

  /* remote crypt key */
  dbKey=GWEN_DB_GetGroup(dbCurr, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			 "bankKeys/cryptKey");
  if (dbKey) {
    GWEN_CRYPTKEY *key;

    key=GWEN_CryptKey_FromDb(dbKey);
    if (key) {
      AH_Medium_SetRemoteCryptKeySpec(m, GWEN_CryptKey_GetKeySpec(key));
      mrdh->remoteCryptKey=key;
    }
  }

  mrdh->selected=1;
  return 0;
}



int AH_MediumRSACard_CreateContext(AH_MEDIUM *m,
				   int country,
				   const char *bankId,
				   const char *userId){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbCurr;
  GWEN_DB_NODE *dbFreeContext;
  int freeIdx;
  int i;
  int lcountry;
  const char *lbankId;
  const char *luserId;


  assert(m);
  assert(bankId);
  assert(userId);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mrdh->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }
  assert(mrdh->dbContexts);

  freeIdx=-1;
  dbFreeContext=0;
  i=0;
  dbCurr=GWEN_DB_FindFirstGroup(mrdh->dbContexts, "context");
  while(dbCurr) {
    lcountry=GWEN_DB_GetIntValue(dbCurr, "country", 0, 280);
    lbankId=GWEN_DB_GetCharValue(dbCurr, "bankCode", 0, 0);
    luserId=GWEN_DB_GetCharValue(dbCurr, "userId", 0, 0);
    if ((lcountry==0 || lbankId==0) && freeIdx==-1) {
      freeIdx=i;
      dbFreeContext=dbCurr;
    }
    if ((lcountry && country && lcountry==country) &&
	(lbankId && bankId && strcasecmp(lbankId, bankId)==0) &&
	(luserId && userId && strcasecmp(luserId, userId)==0)) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "Found context \"%d/%s/%s\"",
	       lcountry, lbankId, luserId);
      break;
    }
    dbCurr=GWEN_DB_FindNextGroup(dbCurr, "context");
    i++;
  } /* while */

  if (dbCurr) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context \"%d/%s/%s\" already exists",
	      country, bankId, userId);
    return -1;
  }

  if (freeIdx==-1) {
    if (i>AH_MEDIUM_RSACARD_MAXCONTEXT) {
      DBG_ERROR(AQHBCI_LOGDOMAIN, "All contexts in use");
      return -1;
    }
    dbFreeContext=GWEN_DB_GetGroup(mrdh->dbContexts,
				   GWEN_PATH_FLAGS_CREATE_GROUP,
				   "context");
  }
  assert(dbFreeContext);

  GWEN_DB_SetIntValue(dbFreeContext, GWEN_DB_FLAGS_DEFAULT,
                      "country", country);
  GWEN_DB_SetCharValue(dbFreeContext, GWEN_DB_FLAGS_DEFAULT,
		       "bankCode", bankId);
  GWEN_DB_SetCharValue(dbFreeContext, GWEN_DB_FLAGS_DEFAULT,
		       "userId", userId);
  GWEN_DB_DeleteGroup(dbFreeContext, "userKeys");
  GWEN_DB_DeleteGroup(dbFreeContext, "bankKeys");

  return freeIdx;
}



int AH_MediumRSACard_ReadContext(AH_MEDIUM *m,
				 int idx,
				 int *country,
				 GWEN_BUFFER *bankId,
				 GWEN_BUFFER *userId,
				 GWEN_BUFFER *server,
				 int *port){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbT;
  const char *s;
  int i;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mrdh->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  i=0;
  dbT=GWEN_DB_FindFirstGroup(mrdh->dbContexts, "context");
  while(dbT) {
    if (i==idx)
      break;
    dbT=GWEN_DB_FindNextGroup(dbT, "context");
    i++;
  } /* while */

  if (!dbT) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context %d not found", idx);
    return -1;
  }

  if (country)
    *country=GWEN_DB_GetIntValue(dbT, "country", 0, 280);
  if (bankId) {
    s=GWEN_DB_GetCharValue(dbT, "bankCode", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(bankId, s);
  }

  if (userId) {
    s=GWEN_DB_GetCharValue(dbT, "userId", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(userId, s);
  }

  if (server) {
    s=GWEN_DB_GetCharValue(dbT, "comAddress", 0, 0);
    if (s)
      GWEN_Buffer_AppendString(server, s);
  }

  if (port) {
    i=GWEN_DB_GetIntValue(dbT, "comService", 0, -1);
    if (i==2)
      *port=3000;
    else if (i==3)
      *port=443;
  }

  return 0;
}



int AH_MediumRSACard_WriteContext(AH_MEDIUM *m,
				  int idx,
				  int country,
				  const char *bankId,
				  const char *userId,
				  const char *server,
				  int port){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbT;
  int i;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!AH_Medium_IsMounted(m)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Medium is not mounted");
    return -1;
  }

  if (!mrdh->dbContexts) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
	      "No contexts, maybe chipcardd is not running?");
    return -1;
  }

  i=0;
  dbT=GWEN_DB_GetFirstGroup(mrdh->dbContexts);
  while(dbT) {
    if (strcasecmp(GWEN_DB_GroupName(dbT), "context")==0) {
      if (i==idx)
	break;
    }
    dbT=GWEN_DB_GetNextGroup(dbT);
    i++;
  } /* while */

  if (!dbT) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Context %d not found", idx);
    return -1;
  }

  if (country)
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"country", country);
  if (bankId)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "bankCode", bankId);
  if (userId)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "userId", userId);
  if (server)
    GWEN_DB_SetCharValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			 "comAddress", server);
  if (port==3000)
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 2);
  else if (port==443) {
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 3);
  }
  else if (port) {
    GWEN_DB_SetIntValue(dbT, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"comService", 0);
  }

  return 0;
}














AH_MEDIUMPROVIDER *rsacard_MediumProvider_new(AH_HBCI *hbci){
  AH_MEDIUMPROVIDER *mp;
  AH_MEDIUMPROVIDER_RSACARD *mprsa;

  mp=AH_MediumProvider_new(hbci, AH_MEDIUM_RSACARD_NAME);
  AH_MediumProvider_SetDeviceType(mp, AH_MediumDeviceChipcard);
  AH_MediumProvider_SetFactoryFn(mp, AH_MediumRSACardProvider_Factory);
  AH_MediumProvider_SetCheckFn(mp, AH_MediumRSACardProvider_Check);

  GWEN_NEW_OBJECT(AH_MEDIUMPROVIDER_RSACARD, mprsa);
  mprsa->client=LC_Client_new("AqHBCI_RSACard", VERSION, 0);
  GWEN_INHERIT_SETDATA(AH_MEDIUMPROVIDER, AH_MEDIUMPROVIDER_RSACARD,
                       mp, mprsa,
                       AH_MediumRSACardProvider_freeData);

  if (LC_Client_ReadConfigFile(mprsa->client, 0)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Error reading configuration.\n");
    AH_MediumProvider_free(mp);
    return 0;
  }

  return mp;
}



void AH_MediumRSACardProvider_freeData(void *bp, void *p) {
  AH_MEDIUMPROVIDER_RSACARD *mprsa;

  mprsa=(AH_MEDIUMPROVIDER_RSACARD*)p;
  LC_Client_free(mprsa->client);
  GWEN_FREE_OBJECT(mprsa);
}



AH_MEDIUM*
AH_MediumRSACardProvider_Factory(AH_MEDIUMPROVIDER *mp,
				 AH_HBCI *hbci,
				 const char *mediumName) {
  return AH_MediumRSACard_new(hbci,
                              mp,
			      mediumName);
}



AH_MEDIUM_CHECKRESULT
AH_MediumRSACardProvider_Check(AH_MEDIUMPROVIDER *mp,
			       AH_HBCI *hbci,
                               GWEN_BUFFER *mediumName) {
  AH_MEDIUMPROVIDER_RSACARD *mprsa;
  LC_CARD *card;
  LC_CLIENT_RESULT res;
  GWEN_DB_NODE *dbCardData;
  const char *p;
  GWEN_TYPE_UINT32 bid;

  assert(mp);
  mprsa=GWEN_INHERIT_GETDATA(AH_MEDIUMPROVIDER,AH_MEDIUMPROVIDER_RSACARD, mp);
  assert(mprsa);

  bid=AB_Banking_ShowBox(AH_HBCI_GetBankingApi(hbci),
                         0,
			 "Accessing Card",
			 "Checking medium, please wait...");

  res=LC_Client_StartWait(mprsa->client, 0, 0);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Wait request timed out");
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }

  card=LC_Client_WaitForNextCard(mprsa->client, 30000);
  if (!card) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "ERROR: No card found");
    LC_Client_StopWait(mprsa->client);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }
  LC_Client_StopWait(mprsa->client);

  if (LC_Starcos_ExtendCard(card)) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "ERROR: Could not extend card as RSA card");
    LC_Card_free(card);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultGenericError;
  }

  res=LC_Card_Open(card);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(AQHBCI_LOGDOMAIN,
	     "Probably not a RSA card (or IO error, check logs)");
    LC_Card_Close(card);
    LC_Card_free(card);
    AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
    return AH_MediumCheckResultWrongMedium;
  }

  dbCardData=LC_Starcos_GetCardDataAsDb(card);
  assert(dbCardData);
  p=GWEN_DB_GetCharValue(dbCardData,
                         "iccsn/cardNumber",
                         0,
                         0);
  if (!p) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "INTERNAL: No card number in card data.");
    abort();
  }

  if (GWEN_Buffer_GetUsedBytes(mediumName)==0) {
    /* set name */
    GWEN_Buffer_AppendString(mediumName, p);
  }
  else {
    /* compare name */
    if (strcasecmp(GWEN_Buffer_GetStart(mediumName), p)!=0) {
      DBG_NOTICE(AQHBCI_LOGDOMAIN, "Wrong medium name");
      LC_Card_Close(card);
      LC_Card_free(card);
      AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
      return AH_MediumCheckResultWrongName;
    }
  }

  LC_Card_Close(card);
  LC_Card_free(card);

  /* do nothing for now */
  AB_Banking_HideBox(AH_HBCI_GetBankingApi(hbci), bid);
  return AH_MediumCheckResultOk;
}



LC_CLIENT *AH_MediumRSACardProvider_GetClient(const AH_MEDIUMPROVIDER *mp){
  AH_MEDIUMPROVIDER_RSACARD *mprsa;

  assert(mp);
  mprsa=GWEN_INHERIT_GETDATA(AH_MEDIUMPROVIDER,
                             AH_MEDIUMPROVIDER_RSACARD,
                             mp);
  return mprsa->client;
}



int AH_MediumRSACard_CreateKeys(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No context selected");
    return AB_ERROR_GENERIC;
  }

  rv=AH_MediumRSACard__EnterEgMode(m, 0);
  if (rv==AB_ERROR_USER1) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "Storing to config file not supported for this function");
    return AB_ERROR_GENERIC;
  }
  else if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  else {
    LC_CLIENT_RESULT res;

    res=LC_Starcos_GenerateKeyPair(mrdh->card, 0x8e, 768);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "GenerateKeyPair");
      return -1;
    }
    res=LC_Starcos_GenerateKeyPair(mrdh->card, 0x8f, 768);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "GenerateKeyPair");
      return -1;
    }
  }

  return 0;
}



int AH_MediumRSACard_ActivateKeys(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No context selected");
    return AB_ERROR_GENERIC;
  }

  rv=AH_MediumRSACard__EnterEgMode(m, 0);
  if (rv==AB_ERROR_USER1) {
    DBG_ERROR(AQHBCI_LOGDOMAIN,
              "Storing to config file not supported for this function");
    return AB_ERROR_GENERIC;
  }
  else if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  else {
    GWEN_KEYSPEC *ks;
    GWEN_CRYPTKEY *key;
    GWEN_DB_NODE *gr;
    LC_CLIENT_RESULT res;

    /* sign key */
    ks=LC_Starcos_GetKeySpec(mrdh->card, 0x80+mrdh->account);
    if (!ks) {
      AH_MediumRSACard__showError(mrdh->card, LC_Client_ResultCmdError,
                                  "GetKeySpec");
      return -1;
    }
  
    if (GWEN_KeySpec_GetStatus(ks)!=0x08) {
      rv=AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                               AB_BANKING_MSG_FLAGS_TYPE_WARN |
                               AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                               AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                               I18N("Key Exists"),
                               I18N("There already is a sign key.\n"
                                    "Do you want to overwrite it ?"
                                    "<html>"
                                    "<p>There already is a sign key.</p>"
                                    "<p>Do you want to overwrite it ?</p>"
                                    "</html>"
                                   ),
                               I18N("Ok"), I18N("Abort"), 0);
      if (rv!=1) {
        GWEN_KeySpec_free(ks);
        return AB_ERROR_USER_ABORT;
      }
      else {
        GWEN_KeySpec_SetStatus(ks, 0x08);
        res=LC_Starcos_SetKeySpec(mrdh->card, 0x80+mrdh->account, ks);
        if (res!=LC_Client_ResultOk) {
          AH_MediumRSACard__showError(mrdh->card, res, "SetKeySpec");
          GWEN_KeySpec_free(ks);
          return -1;
        }
      }
    }

    GWEN_KeySpec_SetStatus(ks, 0x10);
    GWEN_KeySpec_SetKeyName(ks, "S");
    GWEN_KeySpec_SetNumber(ks, 1);
    GWEN_KeySpec_SetVersion(ks, 1);
    res=LC_Starcos_ActivateKeyPair(mrdh->card, 0x8f, 0x80+mrdh->account, ks);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "ActivateKeyPair");
      GWEN_KeySpec_free(ks);
      return -1;
    }

    /* read user keys */
    gr=GWEN_DB_GetGroup(mrdh->dbCurrentContext,
			GWEN_DB_FLAGS_DEFAULT,
			"userKeys");
    assert(gr);

    /* read local sign key */
    DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading local sign key");
    key=LC_Starcos_ReadPublicKey(mrdh->card, 0x80+mrdh->account);
    if (!key) {
      AH_MediumRSACard__showError(mrdh->card, LC_Client_ResultCmdError,
				  "ReadPublicKey");
    }
    else {
      GWEN_DB_NODE *dbKey;
      GWEN_ERRORCODE err;

      GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
      GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
      GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
      dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
			     "signKey");
      err=GWEN_CryptKey_ToDb(key, dbKey, 1);
      if (!GWEN_Error_IsOk(err)) {
	DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
	GWEN_CryptKey_free(key);
	return -1;
      }
      GWEN_CryptKey_free(mrdh->localSignKey);
      mrdh->localSignKey=key;
    }
    GWEN_KeySpec_free(ks);

    /* crypt key */
    ks=LC_Starcos_GetKeySpec(mrdh->card, 0x85+mrdh->account);
    if (!ks) {
      AH_MediumRSACard__showError(mrdh->card, LC_Client_ResultCmdError,
                                  "GetKeySpec");
      return -1;
    }
  
    if (GWEN_KeySpec_GetStatus(ks)!=0x08) {
      rv=AB_Banking_MessageBox(AH_Medium_GetBankingApi(m),
                               AB_BANKING_MSG_FLAGS_TYPE_WARN |
                               AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS |
                               AB_BANKING_MSG_FLAGS_CONFIRM_B1,
                               I18N("Key Exists"),
                               I18N("There already is a crypt key.\n"
                                    "Do you want to overwrite it ?"
                                    "<html>"
                                    "<p>There already is a crypt key.</p>"
                                    "<p>Do you want to overwrite it ?</p>"
                                    "</html>"
                                   ),
                               I18N("Ok"), I18N("Abort"), 0);
      if (rv!=1) {
        GWEN_KeySpec_free(ks);
        return AB_ERROR_USER_ABORT;
      }
      else {
        GWEN_KeySpec_SetStatus(ks, 0x08);
        res=LC_Starcos_SetKeySpec(mrdh->card, 0x85+mrdh->account, ks);
        if (res!=LC_Client_ResultOk) {
          AH_MediumRSACard__showError(mrdh->card, res, "SetKeySpec");
          GWEN_KeySpec_free(ks);
          return -1;
        }
      }
    }

    GWEN_KeySpec_SetStatus(ks, 0x10);
    GWEN_KeySpec_SetKeyName(ks, "V");
    GWEN_KeySpec_SetNumber(ks, 1);
    GWEN_KeySpec_SetVersion(ks, 1);
    res=LC_Starcos_ActivateKeyPair(mrdh->card, 0x8e, 0x85+mrdh->account, ks);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "ActivateKeyPair");
      GWEN_KeySpec_free(ks);
      return -1;
    }

    /* read local crypt key */
    DBG_NOTICE(AQHBCI_LOGDOMAIN, "Reading local crypt key");
    key=LC_Starcos_ReadPublicKey(mrdh->card, 0x85+mrdh->account);
    if (!key) {
      AH_MediumRSACard__showError(mrdh->card, LC_Client_ResultCmdError,
				  "ReadPublicKey");
    }
    else {
      GWEN_DB_NODE *dbKey;
      GWEN_ERRORCODE err;

      GWEN_CryptKey_SetStatus(key, GWEN_KeySpec_GetStatus(ks));
      GWEN_CryptKey_SetNumber(key, GWEN_KeySpec_GetNumber(ks));
      GWEN_CryptKey_SetVersion(key, GWEN_KeySpec_GetVersion(ks));
      dbKey=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
			     "cryptKey");
      err=GWEN_CryptKey_ToDb(key, dbKey, 1);
      if (!GWEN_Error_IsOk(err)) {
	DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
	GWEN_CryptKey_free(key);
	return -1;
      }
      GWEN_CryptKey_free(mrdh->localCryptKey);
      mrdh->localCryptKey=key;
    }
    GWEN_KeySpec_free(ks);

  }


  return 0;
}


GWEN_CRYPTKEY *AH_MediumRSACard_GetLocalPubSignKey(AH_MEDIUM *m){
  GWEN_CRYPTKEY *key;
  GWEN_DB_NODE *db;
  GWEN_ERRORCODE err;
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->localSignKey) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "No key");
    return 0;
  }
  db=GWEN_DB_Group_new("key");
  err=GWEN_CryptKey_ToDb(mrdh->localSignKey, db, 1);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_DB_Group_free(db);
    return 0;
  }

  key=GWEN_CryptKey_FromDb(db);
  if (!key) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not create key from previous export");
    GWEN_DB_Group_free(db);
    return 0;
  }

  GWEN_DB_Group_free(db);
  return key;
}



GWEN_CRYPTKEY *AH_MediumRSACard_GetLocalPubCryptKey(AH_MEDIUM *m){
  GWEN_CRYPTKEY *key;
  GWEN_DB_NODE *db;
  GWEN_ERRORCODE err;
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->localCryptKey) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "No key");
    return 0;
  }
  db=GWEN_DB_Group_new("key");
  err=GWEN_CryptKey_ToDb(mrdh->localCryptKey, db, 1);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
    GWEN_DB_Group_free(db);
    return 0;
  }
  key=GWEN_CryptKey_FromDb(db);
  if (!key) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "Could not create key from previous export");
    GWEN_DB_Group_free(db);
    return 0;
  }
  GWEN_DB_Group_free(db);

  return key;
}


int AH_MediumRSACard__EnterEgMode(AH_MEDIUM *m, int allowConfig) {
  AH_MEDIUM_RSACARD *mrdh;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (mrdh->hasEgPin==0) {
    int reallyEnter;

    if (mrdh->storeToCard==-1 && allowConfig) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "Storing to config file only");
      return AB_ERROR_USER1;
    }

    reallyEnter=0;

    rv=AB_Banking_MessageBox
      (AH_Medium_GetBankingApi(m),
       AB_BANKING_MSG_FLAGS_TYPE_WARN |
       AB_BANKING_MSG_FLAGS_CONFIRM_B2 |
       AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS,
       I18N("Modify Card Data"),
       I18N("The application requests a modification of chip card\n"
            "data.\n"
            "This would require a special PIN to be entered.\n"
            "However, some banks provide you with a personalized\n"
            "card for which this PIN is intentionally unknown to you.\n"
            "In such a case the data can be written to the\n"
            "configuration file only.\n"
            "If you did change the special PIN (which is very \n"
            "unlikely) you can enter it later.\n"
            "\n"
            "How do you want to store the changes ?\n"
            /* same message in HTML */
            "<HTML>"
            "<p>"
            "The application requests a modification of chip card data."
            "</p>"
            "<p>"
            "This would require a special PIN to be entered."
            "</p>"
            "<p>"
            "However, some banks provide you with a personalized card "
            "for which this PIN is intentionally unknown to you."
            "</p>"
            "<p>"
            "In such a case the data can be written to the configuration "
            "file only."
            "</p>"
            "<p>"
            "How do you want to store the changes ?\n"
            "</p>"
            "</HTML>"
           ),
       allowConfig?I18N("Configuration file"):I18N("Chipcard"),
       allowConfig?I18N("Chipcard"):I18N("Abort"),
       allowConfig?I18N("Abort"):0
      );

    if (allowConfig) {
      if (rv==1)
        reallyEnter=0;
      else if (rv==2)
        reallyEnter=1;
      else {
        DBG_WARN(AQHBCI_LOGDOMAIN, "Aborted by user.");
        return AB_ERROR_USER_ABORT;
      }
    }
    else {
      if (rv==1)
        reallyEnter=1;
      else {
        DBG_WARN(AQHBCI_LOGDOMAIN, "Aborted by user.");
        return AB_ERROR_USER_ABORT;
      }
    }

    if (reallyEnter==0) {
      DBG_INFO(AQHBCI_LOGDOMAIN, "Storing to config file only");
      mrdh->storeToCard=1;
      return AB_ERROR_USER1;
    }
    else {
      LC_CLIENT_RESULT res;
  
      rv=AB_Banking_MessageBox
        (AH_Medium_GetBankingApi(m),
         AB_BANKING_MSG_FLAGS_TYPE_WARN |
         AB_BANKING_MSG_FLAGS_CONFIRM_B2 |
         AB_BANKING_MSG_FLAGS_SEVERITY_DANGEROUS,
         I18N("Choose Device PIN"),
         I18N("The device PIN per default is the serial number of the card.\n"
              "However, if you did change the special PIN (which is very \n"
              "unlikely) you can enter it. Otherwise the default PIN should\n"
              "be used.\n"
              "Please choose the PIN entry method:\n"
              /* same message in HTML */
              "<HTML>"
              "<p>The device PIN per default is the serial number of the "
              "card.</p>"
              "<p>However, if you did change the special PIN (which is very "
              "unlikely) you can enter it. Otherwise the default PIN should "
              "be used.</p>"
              "<p>Please choose the PIN entry method:</p>"
              "</HTML>"
             ),
         I18N("Default PIN"),
         I18N("Enter PIN"),
         I18N("Abort")
        );
      if (rv!=1 && rv!=2) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Aborted by user.");
        return AB_ERROR_USER_ABORT;
      }
      res=AH_MediumRSACard__enterPin(m, mrdh->card, 0x91, (rv==1));
      if (res!=LC_Client_ResultOk) {
        DBG_ERROR(AQHBCI_LOGDOMAIN, "Error in PIN verification.");
        return -1;
      }
      mrdh->hasEgPin=1;
      mrdh->storeToCard=-1;
    }
  }

  return 0;
}



GWEN_CRYPTKEY *AH_MediumRSACard_GetPubSignKey(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No user selected");
    return 0;
  }
  if (mrdh->remoteSignKey)
    return GWEN_CryptKey_dup(mrdh->remoteSignKey);
  else
    return 0;
}



int AH_MediumRSACard_SetPubSignKey(AH_MEDIUM *m,
                                   const GWEN_CRYPTKEY *key){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbKey;
  GWEN_ERRORCODE err;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  assert(key);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No user selected");
    return AB_ERROR_GENERIC;
  }
  assert(mrdh->dbCurrentContext);
  assert(mrdh->account);

  rv=AH_MediumRSACard__EnterEgMode(m, 1);
  if (rv==AB_ERROR_USER1) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Storing to config file only");
  }
  else if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  else {
    int kid;
    LC_CLIENT_RESULT res;

    /* write public key */
    kid=0x90+mrdh->account;
    res=LC_Starcos_WritePublicKey(mrdh->card, kid, key);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "WritePublicKey");
      return AB_ERROR_GENERIC;
    }
  }

  GWEN_CryptKey_free(mrdh->remoteSignKey);
  mrdh->remoteSignKey=0;

  dbKey=GWEN_DB_GetGroup(mrdh->dbCurrentContext, GWEN_DB_FLAGS_DEFAULT,
			 "bankKeys");
  assert(dbKey);
  dbKey=GWEN_DB_GetGroup(dbKey, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
			 "signKey");
  assert(dbKey);

  err=GWEN_CryptKey_ToDb(key, dbKey, 1);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
    return AB_ERROR_GENERIC;
  }

  mrdh->remoteSignKey=GWEN_CryptKey_dup(key);

  return 0;
}



GWEN_CRYPTKEY *AH_MediumRSACard_GetPubCryptKey(AH_MEDIUM *m){
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No user selected");
    return 0;
  }

  if (mrdh->remoteCryptKey)
    return GWEN_CryptKey_dup(mrdh->remoteCryptKey);
  else
    return 0;
}



int AH_MediumRSACard_SetPubCryptKey(AH_MEDIUM *m,
                                    const GWEN_CRYPTKEY *key){
  AH_MEDIUM_RSACARD *mrdh;
  GWEN_DB_NODE *dbKey;
  GWEN_ERRORCODE err;
  int rv;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  assert(key);

  if (!mrdh->selected) {
    DBG_ERROR(AQHBCI_LOGDOMAIN, "No user selected");
    return AB_ERROR_GENERIC;
  }

  assert(mrdh->dbCurrentContext);
  assert(mrdh->account);

  rv=AH_MediumRSACard__EnterEgMode(m, 1);
  if (rv==AB_ERROR_USER1) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "Storing to config file only");
  }
  else if (rv) {
    DBG_INFO(AQHBCI_LOGDOMAIN, "here");
    return rv;
  }
  else {
    int kid;
    LC_CLIENT_RESULT res;

    /* write public key */
    kid=0x95+mrdh->account;
    res=LC_Starcos_WritePublicKey(mrdh->card, kid, key);
    if (res!=LC_Client_ResultOk) {
      AH_MediumRSACard__showError(mrdh->card, res, "WritePublicKey");
      return AB_ERROR_GENERIC;
    }
  }

  GWEN_CryptKey_free(mrdh->remoteCryptKey);
  mrdh->remoteCryptKey=0;

  assert(mrdh->dbCurrentContext);
  dbKey=GWEN_DB_GetGroup(mrdh->dbCurrentContext, GWEN_DB_FLAGS_DEFAULT,
			 "bankKeys");
  assert(dbKey);
  dbKey=GWEN_DB_GetGroup(dbKey, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
			 "cryptKey");
  assert(dbKey);

  err=GWEN_CryptKey_ToDb(key, dbKey, 1);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(AQHBCI_LOGDOMAIN, err);
    return AB_ERROR_GENERIC;
  }

  mrdh->remoteCryptKey=GWEN_CryptKey_dup(key);

  return 0;
}



int AH_MediumRSACard_toDb(const AH_MEDIUM *m, GWEN_DB_NODE *db){
  GWEN_DB_NODE *dbT;
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  if (mrdh->dbContexts) {
    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_OVERWRITE_GROUPS, "contexts");
    assert(dbT);
    GWEN_DB_AddGroupChildren(dbT, mrdh->dbContexts);
  }
  return 0;
}



int AH_MediumRSACard_fromDb(AH_MEDIUM *m, GWEN_DB_NODE *db){
  GWEN_DB_NODE *dbT;
  AH_MEDIUM_RSACARD *mrdh;

  assert(m);
  mrdh=GWEN_INHERIT_GETDATA(AH_MEDIUM, AH_MEDIUM_RSACARD, m);
  assert(mrdh);

  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "contexts");
  if (dbT) {
    GWEN_DB_Group_free(mrdh->dbContexts);
    mrdh->dbContexts=GWEN_DB_Group_dup(dbT);
  }
  return 0;
}









